@@ -14,11 +14,13 @@ try:
 except ImportError:
     from typing_extensions import Literal
 
-import cltoolbox
 import pandas as pd
 import typic
+from cltoolbox import Program
 from cltoolbox.rst_text_formatter import RSTHelpFormatter
-from tstoolbox import tsutils
+from toolbox_utils import tsutils
+
+program = Program("hspfbintoolbox", 0.0)
 
 code2intervalmap = {5: "yearly", 4: "monthly", 3: "daily", 2: "bivl"}
 
@@ -34,7 +36,7 @@ _LOCAL_DOCSTRINGS = {
 }
 
 
-def tupleMatch(a, b):
+def tuple_match(a, b):
     """Part of partial ordered matching.
     See http://stackoverflow.com/a/4559604
     """
@@ -43,24 +45,56 @@ def tupleMatch(a, b):
     )
 
 
-def tupleCombine(a, b):
+def tuple_combine(a, b):
     """Part of partial ordered matching.
     See http://stackoverflow.com/a/4559604
     """
     return tuple(i is None and j or i for i, j in zip(a, b))
 
 
-def tupleSearch(findme, haystack):
+def tuple_search(findme, haystack):
     """Partial ordered matching with 'None' as wildcard
     See http://stackoverflow.com/a/4559604
     """
     return [
-        (i, tupleCombine(findme, h))
+        (i, tuple_combine(findme, h))
         for i, h in enumerate(haystack)
-        if tupleMatch(findme, h)
+        if tuple_match(findme, h)
     ]
 
 
+def _range_to_numlist(rangestr):
+    numlist = []
+    subranges = rangestr.split("+")
+    for sub in subranges:
+        try:
+            if ":" not in sub:
+                num = int(sub)
+                numlist.append(num)
+            else:
+                ends = sub.split(":")
+                if len(ends) != 2:
+                    raise ValueError()
+                rstart = int(ends[0])
+                rend = int(ends[1]) + 1
+                if rstart > rend:
+                    raise ValueError()
+                numlist.extend(iter(range(rstart, rend)))
+        except ValueError as exc:
+            raise ValueError(
+                tsutils.error_wrapper(
+                    f"""
+                    Invalid range specification in '{sub}' of the '{rangestr}'.
+                    The correct syntax is one or more integers or
+                    colon-delimited range groups such as "99", "1:2", or
+                    "101:120", with multiple groups connected by "+" signs.
+                    Example: "1:4+16:22+30"
+                    """
+                )
+            ) from exc
+    return numlist
+
+
 def _get_data(binfilename, interval="daily", labels=None, catalog_only=True):
     """Underlying function to read from the binary file.  Used by
     'extract', 'catalog'.
@@ -112,143 +146,203 @@ def _get_data(binfilename, interval="daily", labels=None, catalog_only=True):
     except AttributeError:
         intervalcode = None
 
-    # Fix up and test the labels - could be in it's own function
-    if isinstance(labels, str):
-        labels = labels.split(" ")
-    labels = tsutils.flatten(labels)
+    # convert label tuples to lists
+    labels = list(labels)
+
+    # turn into a list of lists
     nlabels = []
     for label in labels:
         if isinstance(label, str):
             nlabels.append(label.split(","))
         else:
             nlabels.append(label)
-    labels = tsutils.flatten(nlabels)
-    labels = [[a, b, c, d] for a, b, c, d in zip(*[iter(labels)] * 4)]
+    labels = nlabels
 
+    # Check the list members for valid values
     for label in labels:
-        words = label
-        if len(words) != 4:
+        if len(label) != 4:
             raise ValueError(
                 tsutils.error_wrapper(
                     f"""
-The label '{label}' has the wrong number of entries.
-"""
+                    The label '{label}' has the wrong number of entries.
+                    """
                 )
             )
 
-        words = [None if i == "" else i for i in words]
+        # replace empty fields with None
+        words = [None if i == "" else i for i in label]
 
+        # first word must be a valid operation type or None
         if words[0] is not None:
+            # force uppercase before comparison
             words[0] = words[0].upper()
-            if words[0] not in testem.keys():
+            if words[0] not in testem:
                 raise ValueError(
                     tsutils.error_wrapper(
                         f"""
-Operation type must be one of 'PERLND', 'IMPLND', 'RCHRES', or 'BMPRAC',
-or missing (to get all) instead of {words[0]}.
-"""
+                        Operation type must be one of 'PERLND', 'IMPLND',
+                        'RCHRES', or 'BMPRAC', or missing (to get all) instead
+                        of {words[0]}.
+                        """
                     )
                 )
 
+        # second word must be integer 1-999 or None or range to parse
         if words[1] is not None:
             try:
                 words[1] = int(words[1])
-                if words[1] < 1 or words[1] > 999:
-                    raise ValueError()
-            except (ValueError, TypeError):
-                raise ValueError(
-                    tsutils.error_wrapper(
-                        f"""
-The land use element must be an integer from 1 to 999 inclusive,
-instead of {words[1]}.
-"""
+                luelist = [words[1]]
+            except ValueError:
+                luelist = _range_to_numlist(words[1])
+            for luenum in luelist:
+                if luenum < 1 or luenum > 999:
+                    raise ValueError(
+                        tsutils.error_wrapper(
+                            f"""
+                            The land use element must be an integer from 1 to
+                            999 inclusive, instead of {luenum}.
+                            """
+                        )
                     )
-                )
+        else:
+            luelist = [None]
 
+        # third word must be a valid group name or None
         if words[2] is not None:
             words[2] = words[2].upper()
             if words[2] not in testem[words[0]]:
                 raise ValueError(
                     tsutils.error_wrapper(
                         f"""
-The {words[0]} operation type only allows the variable groups:
-{testem[words[0]][:-1]},
-instead you gave {words[2]}.
-"""
+                        The {words[0]} operation type only allows the variable
+                        groups: {testem[words[0]][:-1]},
+                        instead you gave {words[2]}.
+                        """
                     )
                 )
 
+        # fourth word is currently not checked - assumed to be a variable name
+        # if not, it will simply never be found in the file, so ok
+        # but no warning for the user - add check?
+
+        # add interval code as fifth word in list
         words.append(intervalcode)
-        lablist.append(words)
 
-    with open(binfilename, "rb") as fl:
+        # add to new list of checked and expanded lists
+        for luenum in luelist:
+            words[1] = luenum
+            lablist.append(list(words))
 
-        mindate = datetime.datetime.max
-        maxdate = datetime.datetime.min
+    # Now read through the binary file and collect the data matching the labels
+    with open(binfilename, "rb") as binfp:
 
         labeltest = set()
         vnames = {}
         ndates = set()
-        rectype = 0
-        fl.read(1)
+        # read first byte - must be hex FD (decimal 253) for valid file.
+        magicbyte = binfp.read(1)
+        if magicbyte != b"\xfd":
+            # not a valid HSPF binary file
+            raise ValueError(
+                tsutils.error_wrapper(
+                    f"""
+                    {binfilename} is not a valid HSPF binary output file
+                    (.hbn),  The first byte must be FD hexadecimal, but it was
+                    {magicbyte}.
+                    """
+                )
+            )
+
+        # loop through each record
         while True:
+            # reinitialize counter for record length - used to compute skip at
+            # end
+            recpos = 0
+
+            # read first four bytes to get record length bitfield
             try:
-                reclen1, reclen2, reclen3, reclen = struct.unpack("4B", fl.read(4))
+                reclen1, reclen2, reclen3, reclen = struct.unpack("4B", binfp.read(4))
+                recpos += 4
             except struct.error:
                 # End of file.
                 break
 
-            rectype, optype, lue, section = struct.unpack("I8sI8s", fl.read(24))
+            # get record leader - next 24 bytes
+            rectype, optype, lue, group = struct.unpack("I8sI8s", binfp.read(24))
+            recpos += 24
 
+            # clean up
             rectype = int(rectype)
             lue = int(lue)
             optype = optype.strip()
-            section = section.strip()
+            group = group.strip()
 
-            slen = 0
             if rectype == 0:
+                # header record - collect variable names for this
+                # operation and group
+
+                # parse reclen bitfield to get actual remaining length
+                # the " - 24 " subtracts the 24 bytes already read
                 reclen1 = int(reclen1 / 4)
                 reclen2 = reclen2 * 64 + reclen1
                 reclen3 = reclen3 * 16384 + reclen2
                 reclen = reclen * 4194304 + reclen3 - 24
+
+                # loop through rest of record
+                slen = 0
                 while slen < reclen:
-                    length = struct.unpack("I", fl.read(4))[0]
-                    slen = slen + length + 4
-                    variable_name = struct.unpack(f"{length}s", fl.read(length))[0]
-                    vnames.setdefault((lue, section), []).append(variable_name)
+                    # read single 4B word for length of next variable name
+                    length = struct.unpack("I", binfp.read(4))[0]
+
+                    # read the variable name
+                    variable_name = struct.unpack(f"{length}s", binfp.read(length))[0]
+
+                    # add variable name to the set for this operation
+                    # why a set instead of a list? There should never be
+                    # a duplicate anyway
+                    vnames.setdefault((lue, group), []).append(variable_name)
+
+                    # update how far along the record we are
+                    slen += length + 4
+                    recpos += length + 4
 
             elif rectype == 1:
                 # Data record
-                numvals = len(vnames[(lue, section)])
+
+                # record should contain a value for each variable name for this
+                # operation and group
+                numvals = len(vnames[(lue, group)])
 
                 (_, level, year, month, day, hour, minute) = struct.unpack(
-                    "7I", fl.read(28)
+                    "7I", binfp.read(28)
                 )
+                recpos += 28
 
-                vals = struct.unpack(f"{numvals}f", fl.read(4 * numvals))
+                vals = struct.unpack(f"{numvals}f", binfp.read(4 * numvals))
+                recpos += 4 * numvals
 
                 delta = datetime.timedelta(hours=0)
                 if hour == 24:
                     hour = 0
-                    # if intervalcode == 2:
-                    #    delta = datetime.timedelta(hours=24)
 
                 ndate = datetime.datetime(year, month, day, hour, minute) + delta
 
-                for i, vname in enumerate(vnames[(lue, section)]):
+                #  Go through labels to see if these values need to be
+                #  collected
+                for i, vname in enumerate(vnames[(lue, group)]):
                     tmpkey = (
                         optype.decode("ascii"),
                         lue,
-                        section.decode("ascii"),
+                        group.decode("ascii"),
                         vname.decode("ascii"),
                         level,
                     )
 
-                    for lb in lablist:
-                        res = tupleSearch(tmpkey, [lb])
+                    for lbl in lablist:
+                        res = tuple_search(tmpkey, [lbl])
                         if not res:
                             continue
-                        labeltest.add(tuple(lb))
+                        labeltest.add(tuple(lbl))
                         nres = res[0][1]
                         ndates.add(ndate)
                         if catalog_only is False:
@@ -257,42 +351,48 @@ instead you gave {words[2]}.
                         else:
                             collect_dict[nres] = level
             else:
-                fl.seek(-31, 1)
+                # there was a problem with unexpected record length
+                # back up almost all the way and try again
+                binfp.seek(-31, 1)
 
-            # The following should be 1 or 2, but I don't know how to calculate
-            # it, so I just use that the next value read is 'rectype' which
-            # must be 0 or 1, and if not means that the fl.read(2) should have
-            # been fl.read(1) and rewind the correct amount.  This is a hack,
-            # but it works. This is done at the beginning of the loop, so it
-            # should be fine.
-            fl.read(2)
+            # calculate and skip to the end of the variable-length back pointer
+            reccnt = recpos * 4 + 1
+            if reccnt >= 256**2:
+                skbytes = 3
+            elif reccnt >= 256:
+                skbytes = 2
+            else:
+                skbytes = 1
+            binfp.read(skbytes)
 
     if not collect_dict:
         raise ValueError(
             tsutils.error_wrapper(
                 f"""
-The label specifications below matched no records in the binary file.
+                The label specifications below matched no records in the binary
+                file.
 
-{lablist}
-"""
+                {lablist}
+                """
             )
         )
 
     ndates = sorted(list(ndates))
 
     if catalog_only is False:
-        for lb in lablist:
-            if tuple(lb) not in labeltest:
+        for lbl in lablist:
+            if tuple(lbl) not in labeltest:
                 sys.stderr.write(
                     tsutils.error_wrapper(
                         f"""
-Warning: The label '{lb}' matched no records in the binary file.
-"""
+                        Warning: The label '{lbl}' matched no records in the
+                        binary file.
+                        """
                     )
                 )
     else:
         for key in collect_dict:
-            if key[4] == 2:
+            if key[4] == 2:  # timestep is bivl
                 delta = ndates[1] - ndates[0]
             else:
                 delta = code2freqmap[key[4]]
@@ -304,9 +404,9 @@ Warning: The label '{lb}' matched no records in the binary file.
     return ndates, collect_dict
 
 
-@cltoolbox.command("extract", formatter_class=RSTHelpFormatter)
+@program.command("extract", formatter_class=RSTHelpFormatter)
 @tsutils.doc(tsutils.merge_dicts(tsutils.docstrings, _LOCAL_DOCSTRINGS))
-def extract_cli(
+def _extract_cli(
     hbnfilename, interval, start_date=None, end_date=None, sort_columns=False, *labels
 ):
     r"""Prints out data to the screen from a HSPF binary output file.
@@ -323,11 +423,11 @@ def extract_cli(
 
     labels: str
         The remaining arguments uniquely identify a time-series in the
-        binary file.  The format is 'OPERATIONTYPE,ID,VARIABLE_GROUP,VARIABLE'.
+        binary file.  The format is 'OPERATIONTYPE,ID,VARIABLEGROUP,VARIABLE'.
 
         For example: 'PERLND,101,PWATER,UZS IMPLND,101,IWATER,RETS'
 
-        Leaving a section without an entry will wildcard that
+        Leaving a section without an entry will wild card that
         specification.  To get all the PWATER variables for PERLND 101 the
         label would read:
 
@@ -337,13 +437,35 @@ def extract_cli(
 
         'PERLND,,,TAET'
 
-        Note that there are spaces ONLY between label specifications.
+        Note that there are spaces ONLY between label specifications not within
+        the labels themselves.
 
         OPERATIONTYE can be PERLND, IMPLND, RCHRES, and BMPRAC.
 
-        ID is specified in the UCI file.
+        ID is the operation type identification number specified in the UCI
+        file. These numbers must be in the range 1-999.
 
-        VARIABLE_GROUP depends on OPERATIONTYPE where::
+        Here, the user can specify
+
+            - a single ID number to match
+            - no entry, matching any operation ID number
+            - a range, specified as any combination of simple integers and
+              groups of integers marked as "start:end", with multiple allowed
+              sub-ranges separated by the "+" sign.
+
+        Examples:
+
+            +-----------------------+-------------------------------+
+            | Label ID              | Expands to:                   |
+            +=======================+===============================+
+            | 1:10                  | 1,2,3,4,5,6,7,8,9,10          |
+            +-----------------------+-------------------------------+
+            | 101:119+221:239       | 101,102..119,221,221,...239   |
+            +-----------------------+-------------------------------+
+            | 3:5+7                 | 3,4,5,7                       |
+            +-----------------------+-------------------------------+
+
+        VARIABLEGROUP depends on OPERATIONTYPE where::
 
             if OPERATIONTYPE is PERLND then VARIABLEGROUP can be one of
                 'ATEMP', 'SNOW', 'PWATER', 'SEDMNT', 'PSTEMP', 'PWTGAS',
@@ -356,10 +478,14 @@ def extract_cli(
                 'HYDR', 'CONS', 'HTRCH', 'SEDTRN', 'GQUAL', 'OXRX', 'NUTRX',
                 'PLANK', 'PHCARB', 'INFLOW', 'OFLOW', 'ROFLOW'
 
-            if OPERATIONTYPE is BMPRAC then there is no VARIABLEGROUP and you
+            if OPERATIONTYPE is BMPRAC then VARIABLEGROUP is not used and you
             have to leave VARIABLEGROUP as a wild card.  For example,
             'BMPRAC,875,,RMVOL'.
 
+        The Time Series Catalog in the HSPF Manual lists all of the variables
+        in each of these VARIABLEGROUPs.  For BMPRAC, all of the variables in
+        all Groups in the Catalog are available in the unnamed (blank) Group.
+
     ${start_date}
     ${end_date}
     sort_columns:
@@ -394,20 +520,16 @@ def extract(
         raise ValueError(
             tsutils.error_wrapper(
                 f"""
-The "interval" argument must be one of "bivl",
-"daily", "monthly", or "yearly".  You supplied
-"{interval}".
-"""
+                The "interval" argument must be one of "bivl", "daily",
+                "monthly", or "yearly".  You supplied "{interval}".
+                """
             )
         )
 
     index, data = _get_data(hbnfilename, interval, labels, catalog_only=False)
-    index = sorted(list(index))
     skeys = list(data.keys())
     if sort_columns:
         skeys.sort(key=lambda tup: tup[1:])
-    else:
-        skeys.sort()
 
     result = pd.DataFrame(
         pd.concat(
@@ -428,9 +550,9 @@ The "interval" argument must be one of "bivl",
     return result
 
 
-@cltoolbox.command("catalog", formatter_class=RSTHelpFormatter)
+@program.command("catalog", formatter_class=RSTHelpFormatter)
 @tsutils.doc(tsutils.merge_dicts(tsutils.docstrings, _LOCAL_DOCSTRINGS))
-def catalog_cli(hbnfilename, tablefmt="simple", header="default"):
+def _catalog_cli(hbnfilename, tablefmt="simple", header="default"):
     """
     Prints out a catalog of data sets in the binary file.
 
@@ -446,7 +568,9 @@ def catalog_cli(hbnfilename, tablefmt="simple", header="default"):
     """
     if header == "default":
         header = ["LUE", "LC", "GROUP", "VAR", "TC", "START", "END", "TC"]
-    tsutils.printiso(catalog(hbnfilename), tablefmt=tablefmt, headers=header)
+    tsutils.printiso(
+        catalog(hbnfilename), tablefmt=tablefmt, headers=header, showindex=False
+    )
 
 
 @typic.al
@@ -461,7 +585,7 @@ def catalog(hbnfilename: str):
     return [cat + catlog[cat] + (code2intervalmap[cat[-1]],) for cat in catkeys]
 
 
-@cltoolbox.command()
+@program.command()
 def about():
     """Display version number and system information."""
     tsutils.about(__name__)
@@ -474,7 +598,7 @@ def main():
         import functiontrace
 
         functiontrace.trace()
-    cltoolbox.main()
+    program()
 
 
 if __name__ == "__main__":
